//
//  TrendingHashtagsViewController.swift
//  Tusker
//
//  Created by Shadowfacts on 2/6/21.
//  Copyright © 2021 Shadowfacts. All rights reserved.
//

import UIKit
import Pachyderm
import WebURLFoundationExtras

class TrendingHashtagsViewController: UIViewController {

    weak var mastodonController: MastodonController!
    
    private var collectionView: UICollectionView!
    private var dataSource: UICollectionViewDiffableDataSource<Section, Item>!
    
    init(mastodonController: MastodonController) {
        self.mastodonController = mastodonController
        
        super.init(nibName: nil, bundle: nil)
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()

        title = NSLocalizedString("Trending Hashtags", comment: "trending hashtags screen title")
        
        view.backgroundColor = .systemGroupedBackground
        
        let config = UICollectionLayoutListConfiguration(appearance: .grouped)
        let layout = UICollectionViewCompositionalLayout.list(using: config)
        collectionView = UICollectionView(frame: view.bounds, collectionViewLayout: layout)
        collectionView.autoresizingMask = [.flexibleWidth, .flexibleHeight]
        collectionView.delegate = self
        collectionView.dragDelegate = self
        view.addSubview(collectionView)
        
        let registration = UICollectionView.CellRegistration<TrendingHashtagCollectionViewCell, Hashtag> { cell, indexPath, hashtag in
            cell.updateUI(hashtag: hashtag)
        }
        
        dataSource = UICollectionViewDiffableDataSource<Section, Item>(collectionView: collectionView) { (collectionView, indexPath, item) in
            switch item {
            case let .tag(hashtag):
                return collectionView.dequeueConfiguredReusableCell(using: registration, for: indexPath, item: hashtag)
            }
        }
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        let request = Client.getTrendingHashtags(limit: 10)
        Task {
            guard let (hashtags, _) = try? await mastodonController.run(request) else {
                return
            }
            var snapshot = NSDiffableDataSourceSnapshot<Section, Item>()
            snapshot.appendSections([.trendingTags])
            snapshot.appendItems(hashtags.map { .tag($0) })
            await dataSource.apply(snapshot)
        }
    }
    
}

extension TrendingHashtagsViewController {
    enum Section {
        case trendingTags
    }
    enum Item: Hashable {
        case tag(Hashtag)
    }
}

extension TrendingHashtagsViewController: UICollectionViewDelegate {
    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        guard let item = dataSource.itemIdentifier(for: indexPath),
              case let .tag(hashtag) = item else {
            return
        }
        
        show(HashtagTimelineViewController(for: hashtag, mastodonController: mastodonController), sender: nil)
    }
    
    func collectionView(_ collectionView: UICollectionView, contextMenuConfigurationForItemAt indexPath: IndexPath, point: CGPoint) -> UIContextMenuConfiguration? {
        guard let item = dataSource.itemIdentifier(for: indexPath),
              case let .tag(hashtag) = item else {
                  return nil
              }
        return UIContextMenuConfiguration(identifier: nil) {
            HashtagTimelineViewController(for: hashtag, mastodonController: self.mastodonController)
        } actionProvider: { (_) in
            UIMenu(children: self.actionsForHashtag(hashtag, sourceView: self.collectionView.cellForItem(at: indexPath)))
        }
    }
}

extension TrendingHashtagsViewController: UICollectionViewDragDelegate {
    func collectionView(_ collectionView: UICollectionView, itemsForBeginning session: UIDragSession, at indexPath: IndexPath) -> [UIDragItem] {
        guard let item = dataSource.itemIdentifier(for: indexPath),
              case let .tag(hashtag) = item,
              let url = URL(hashtag.url) else {
                  return []
              }
        let provider = NSItemProvider(object: url as NSURL)
        if let activity = UserActivityManager.showTimelineActivity(timeline: .tag(hashtag: hashtag.name), accountID: mastodonController.accountInfo!.id) {
            activity.displaysAuxiliaryScene = true
            provider.registerObject(activity, visibility: .all)
        }
        return [UIDragItem(itemProvider: provider)]
    }
}

extension TrendingHashtagsViewController: TuskerNavigationDelegate {
    var apiController: MastodonController! { mastodonController }
}

extension TrendingHashtagsViewController: ToastableViewController {
}

extension TrendingHashtagsViewController: MenuActionProvider {
}
